from airbyte_cdk.sources.file_based.stream.abstract_file_based_stream import AbstractFileBasedStream
from airbyte_cdk.sources.file_based.stream.default_file_based_stream import DefaultFileBasedStream
from airbyte_cdk.sources.file_based.stream.identities_stream import FileIdentitiesStream
from airbyte_cdk.sources.file_based.stream.permissions_file_based_stream import (
    PermissionsFileBasedStream,
)

__all__ = [
    "AbstractFileBasedStream",
    "DefaultFileBasedStream",
    "FileIdentitiesStream",
    "PermissionsFileBasedStream",
]



import hashlib
import json
from typing import Any, Final, Mapping, Optional


class SliceEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if hasattr(obj, "__json_serializable__"):
            return obj.__json_serializable__()

        # Let the base class default method raise the TypeError
        return super().default(obj)


class SliceHasher:
    _ENCODING: Final = "utf-8"

    @classmethod
    def hash(
        cls,
        stream_name: str = "<stream name not provided>",
        stream_slice: Optional[Mapping[str, Any]] = None,
    ) -> int:
        """
        Note that streams partition with the same slicing value but with different names might collapse if stream name is not provided
        """
        if stream_slice:
            try:
                s = json.dumps(stream_slice, sort_keys=True, cls=SliceEncoder)
                hash_input = f"{stream_name}:{s}".encode(cls._ENCODING)
            except TypeError as e:
                raise ValueError(f"Failed to serialize stream slice: {e}")
        else:
            hash_input = stream_name.encode(cls._ENCODING)

        # Use last 8 bytes as 64-bit integer for better distribution
        return int.from_bytes(hashlib.sha256(hash_input).digest()[-8:], "big")



#
# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Mapping, Optional

from airbyte_cdk.sources.declarative.stream_slicers.stream_slicer import StreamSlicer
from airbyte_cdk.sources.types import StreamState


@dataclass
class PartitionRouter(StreamSlicer):
    """
    Base class for partition routers.
    Methods:
        set_parent_state(stream_state): Set the state of the parent streams.
        get_parent_state(): Get the state of the parent streams.
    """

    @abstractmethod
    def set_initial_state(self, stream_state: StreamState) -> None:
        """
        Set the state of the parent streams.

        This method should only be implemented if the slicer is based on some parent stream and needs to read this stream
        incrementally using the state.

        Args:
            stream_state (StreamState): The state of the streams to be set. The expected format is a dictionary that includes
                                        'parent_state' which is a dictionary of parent state names to their corresponding state.
                Example:
                {
                    "parent_state": {
                        "parent_stream_name_1": { ... },
                        "parent_stream_name_2": { ... },
                        ...
                    }
                }
        """

    @abstractmethod
    def get_stream_state(self) -> Optional[Mapping[str, StreamState]]:
        """
        Get the state of the parent streams.

        This method should only be implemented if the slicer is based on some parent stream and needs to read this stream
        incrementally using the state.

        Returns:
            Optional[Mapping[str, StreamState]]: The current state of the parent streams in a dictionary format.
                 The returned format will be:
                 {
                     "parent_stream_name1": {
                         "last_updated": "2023-05-27T00:00:00Z"
                     },
                     "parent_stream_name2": {
                         "last_updated": "2023-05-27T00:00:00Z"
                     }
                 }
        """



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from dataclasses import dataclass
from typing import Any, Mapping, MutableMapping, Optional

from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.streams.checkpoint import Cursor
from airbyte_cdk.sources.streams.checkpoint.per_partition_key_serializer import (
    PerPartitionKeySerializer,
)
from airbyte_cdk.sources.types import Record, StreamSlice, StreamState
from airbyte_cdk.utils import AirbyteTracedException

FULL_REFRESH_COMPLETE_STATE: Mapping[str, Any] = {"__ab_full_refresh_sync_complete": True}


@dataclass
class SubstreamResumableFullRefreshCursor(Cursor):
    def __init__(self) -> None:
        self._per_partition_state: MutableMapping[str, StreamState] = {}
        self._partition_serializer = PerPartitionKeySerializer()

    def get_stream_state(self) -> StreamState:
        return {"states": list(self._per_partition_state.values())}

    def set_initial_state(self, stream_state: StreamState) -> None:
        """
        Set the initial state for the cursors.

        This method initializes the state for each partition cursor using the provided stream state.
        If a partition state is provided in the stream state, it will update the corresponding partition cursor with this state.

        To simplify processing and state management, we do not maintain the checkpointed state of the parent partitions.
        Instead, we are tracking whether a parent has already successfully synced on a prior attempt and skipping over it
        allowing the sync to continue making progress. And this works for RFR because the platform will dispose of this
        state on the next sync job.

        Args:
            stream_state (StreamState): The state of the streams to be set. The format of the stream state should be:
                {
                    "states": [
                        {
                            "partition": {
                                "partition_key": "value_0"
                            },
                            "cursor": {
                                "__ab_full_refresh_sync_complete": True
                            }
                        },
                        {
                            "partition": {
                                "partition_key": "value_1"
                            },
                            "cursor": {},
                        },
                    ]
                }
        """
        if not stream_state:
            return

        if "states" not in stream_state:
            raise AirbyteTracedException(
                internal_message=f"Could not sync parse the following state: {stream_state}",
                message="The state for is format invalid. Validate that the migration steps included a reset and that it was performed "
                "properly. Otherwise, please contact Airbyte support.",
                failure_type=FailureType.config_error,
            )

        for state in stream_state["states"]:
            self._per_partition_state[self._to_partition_key(state["partition"])] = state

    def observe(self, stream_slice: StreamSlice, record: Record) -> None:
        """
        Substream resumable full refresh manages state by closing the slice after syncing a parent so observe is not used.
        """
        pass

    def close_slice(self, stream_slice: StreamSlice, *args: Any) -> None:
        self._per_partition_state[self._to_partition_key(stream_slice.partition)] = {
            "partition": stream_slice.partition,
            "cursor": FULL_REFRESH_COMPLETE_STATE,
        }

    def should_be_synced(self, record: Record) -> bool:
        """
        Unlike date-based cursors which filter out records outside slice boundaries, resumable full refresh records exist within pages
        that don't have filterable bounds. We should always return them.
        """
        return True

    def is_greater_than_or_equal(self, first: Record, second: Record) -> bool:
        """
        RFR record don't have ordering to be compared between one another.
        """
        return False

    def select_state(self, stream_slice: Optional[StreamSlice] = None) -> Optional[StreamState]:
        if not stream_slice:
            raise ValueError("A partition needs to be provided in order to extract a state")

        return self._per_partition_state.get(
            self._to_partition_key(stream_slice.partition), {}
        ).get("cursor")

    def _to_partition_key(self, partition: Mapping[str, Any]) -> str:
        return self._partition_serializer.to_partition_key(partition)

    def _to_dict(self, partition_key: str) -> Mapping[str, Any]:
        return self._partition_serializer.to_partition(partition_key)



from airbyte_cdk.sources.file_based.discovery_policy.abstract_discovery_policy import (
    AbstractDiscoveryPolicy,
)
from airbyte_cdk.sources.file_based.discovery_policy.default_discovery_policy import (
    DefaultDiscoveryPolicy,
)

__all__ = ["AbstractDiscoveryPolicy", "DefaultDiscoveryPolicy"]



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass
from typing import Any, Dict, Mapping

from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader


@dataclass
class InlineSchemaLoader(SchemaLoader):
    """Describes a stream's schema"""

    schema: Dict[str, Any]
    parameters: InitVar[Mapping[str, Any]]

    def get_json_schema(self) -> Mapping[str, Any]:
        return self.schema



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from dataclasses import InitVar, dataclass, field
from datetime import datetime, timedelta
from typing import Any, List, Mapping, MutableMapping, Optional, Union

from airbyte_cdk.sources.declarative.auth.declarative_authenticator import DeclarativeAuthenticator
from airbyte_cdk.sources.declarative.interpolation.interpolated_boolean import InterpolatedBoolean
from airbyte_cdk.sources.declarative.interpolation.interpolated_mapping import InterpolatedMapping
from airbyte_cdk.sources.declarative.interpolation.interpolated_string import InterpolatedString
from airbyte_cdk.sources.message import MessageRepository, NoopMessageRepository
from airbyte_cdk.sources.streams.http.requests_native_auth.abstract_oauth import (
    AbstractOauth2Authenticator,
)
from airbyte_cdk.sources.streams.http.requests_native_auth.oauth import (
    SingleUseRefreshTokenOauth2Authenticator,
)
from airbyte_cdk.utils.datetime_helpers import AirbyteDateTime, ab_datetime_now, ab_datetime_parse


@dataclass
class DeclarativeOauth2Authenticator(AbstractOauth2Authenticator, DeclarativeAuthenticator):
    """
    Generates OAuth2.0 access tokens from an OAuth2.0 refresh token and client credentials based on
    a declarative connector configuration file. Credentials can be defined explicitly or via interpolation
    at runtime. The generated access token is attached to each request via the Authorization header.

    Attributes:
        token_refresh_endpoint (Union[InterpolatedString, str]): The endpoint to refresh the access token
        client_id (Union[InterpolatedString, str]): The client id
        client_secret (Union[InterpolatedString, str]): Client secret
        refresh_token (Union[InterpolatedString, str]): The token used to refresh the access token
        access_token_name (Union[InterpolatedString, str]): THe field to extract access token from in the response
        expires_in_name (Union[InterpolatedString, str]): The field to extract expires_in from in the response
        config (Mapping[str, Any]): The user-provided configuration as specified by the source's spec
        scopes (Optional[List[str]]): The scopes to request
        token_expiry_date (Optional[Union[InterpolatedString, str]]): The access token expiration date
        token_expiry_date_format str: format of the datetime; provide it if expires_in is returned in datetime instead of seconds
        token_expiry_is_time_of_expiration bool: set True it if expires_in is returned as time of expiration instead of the number seconds until expiration
        refresh_request_body (Optional[Mapping[str, Any]]): The request body to send in the refresh request
        refresh_request_headers (Optional[Mapping[str, Any]]): The request headers to send in the refresh request
        grant_type: The grant_type to request for access_token. If set to refresh_token, the refresh_token parameter has to be provided
        message_repository (MessageRepository): the message repository used to emit logs on HTTP requests
    """

    config: Mapping[str, Any]
    parameters: InitVar[Mapping[str, Any]]
    client_id: Optional[Union[InterpolatedString, str]] = None
    client_secret: Optional[Union[InterpolatedString, str]] = None
    token_refresh_endpoint: Optional[Union[InterpolatedString, str]] = None
    refresh_token: Optional[Union[InterpolatedString, str]] = None
    scopes: Optional[List[str]] = None
    token_expiry_date: Optional[Union[InterpolatedString, str]] = None
    _token_expiry_date: Optional[AirbyteDateTime] = field(init=False, repr=False, default=None)
    token_expiry_date_format: Optional[str] = None
    token_expiry_is_time_of_expiration: bool = False
    access_token_name: Union[InterpolatedString, str] = "access_token"
    access_token_value: Optional[Union[InterpolatedString, str]] = None
    client_id_name: Union[InterpolatedString, str] = "client_id"
    client_secret_name: Union[InterpolatedString, str] = "client_secret"
    expires_in_name: Union[InterpolatedString, str] = "expires_in"
    refresh_token_name: Union[InterpolatedString, str] = "refresh_token"
    refresh_request_body: Optional[Mapping[str, Any]] = None
    refresh_request_headers: Optional[Mapping[str, Any]] = None
    grant_type_name: Union[InterpolatedString, str] = "grant_type"
    grant_type: Union[InterpolatedString, str] = "refresh_token"
    message_repository: MessageRepository = NoopMessageRepository()
    profile_assertion: Optional[DeclarativeAuthenticator] = None
    use_profile_assertion: Optional[Union[InterpolatedBoolean, str, bool]] = False

    def __post_init__(self, parameters: Mapping[str, Any]) -> None:
        super().__init__()
        if self.token_refresh_endpoint is not None:
            self._token_refresh_endpoint: Optional[InterpolatedString] = InterpolatedString.create(
                self.token_refresh_endpoint, parameters=parameters
            )
        else:
            self._token_refresh_endpoint = None
        self._client_id_name = InterpolatedString.create(self.client_id_name, parameters=parameters)
        self._client_id = (
            InterpolatedString.create(self.client_id, parameters=parameters)
            if self.client_id
            else self.client_id
        )
        self._client_secret_name = InterpolatedString.create(
            self.client_secret_name, parameters=parameters
        )
        self._client_secret = (
            InterpolatedString.create(self.client_secret, parameters=parameters)
            if self.client_secret
            else self.client_secret
        )
        self._refresh_token_name = InterpolatedString.create(
            self.refresh_token_name, parameters=parameters
        )
        if self.refresh_token is not None:
            self._refresh_token: Optional[InterpolatedString] = InterpolatedString.create(
                self.refresh_token, parameters=parameters
            )
        else:
            self._refresh_token = None
        self.access_token_name = InterpolatedString.create(
            self.access_token_name, parameters=parameters
        )
        self.expires_in_name = InterpolatedString.create(
            self.expires_in_name, parameters=parameters
        )
        self.grant_type_name = InterpolatedString.create(
            self.grant_type_name, parameters=parameters
        )
        self.grant_type = InterpolatedString.create(
            "urn:ietf:params:oauth:grant-type:jwt-bearer"
            if self.use_profile_assertion
            else self.grant_type,
            parameters=parameters,
        )
        self._refresh_request_body = InterpolatedMapping(
            self.refresh_request_body or {}, parameters=parameters
        )
        self._refresh_request_headers = InterpolatedMapping(
            self.refresh_request_headers or {}, parameters=parameters
        )
        try:
            if (
                isinstance(self.token_expiry_date, (int, str))
                and str(self.token_expiry_date).isdigit()
            ):
                self._token_expiry_date = ab_datetime_parse(self.token_expiry_date)
            else:
                self._token_expiry_date = (
                    ab_datetime_parse(
                        InterpolatedString.create(
                            self.token_expiry_date, parameters=parameters
                        ).eval(self.config)
                    )
                    if self.token_expiry_date
                    else ab_datetime_now() - timedelta(days=1)
                )
        except ValueError as e:
            raise ValueError(f"Invalid token expiry date format: {e}")
        self.use_profile_assertion = (
            InterpolatedBoolean(self.use_profile_assertion, parameters=parameters)
            if isinstance(self.use_profile_assertion, str)
            else self.use_profile_assertion
        )
        self.assertion_name = "assertion"

        if self.access_token_value is not None:
            self._access_token_value = InterpolatedString.create(
                self.access_token_value, parameters=parameters
            ).eval(self.config)
        else:
            self._access_token_value = None

        self._access_token: Optional[str] = (
            self._access_token_value if self.access_token_value else None
        )

        if not self.use_profile_assertion and any(
            client_creds is None for client_creds in [self.client_id, self.client_secret]
        ):
            raise ValueError(
                "OAuthAuthenticator configuration error: Both 'client_id' and 'client_secret' are required for the "
                "basic OAuth flow."
            )
        if self.profile_assertion is None and self.use_profile_assertion:
            raise ValueError(
                "OAuthAuthenticator configuration error: 'profile_assertion' is required when using the profile assertion flow."
            )
        if self.get_grant_type() == "refresh_token" and self._refresh_token is None:
            raise ValueError(
                "OAuthAuthenticator configuration error: A 'refresh_token' is required when the 'grant_type' is set to 'refresh_token'."
            )

    def get_token_refresh_endpoint(self) -> Optional[str]:
        if self._token_refresh_endpoint is not None:
            refresh_token_endpoint: str = self._token_refresh_endpoint.eval(self.config)
            if not refresh_token_endpoint:
                raise ValueError(
                    "OAuthAuthenticator was unable to evaluate token_refresh_endpoint parameter"
                )
            return refresh_token_endpoint
        return None

    def get_client_id_name(self) -> str:
        return self._client_id_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_client_id(self) -> str:
        client_id = self._client_id.eval(self.config) if self._client_id else self._client_id
        if not client_id:
            raise ValueError("OAuthAuthenticator was unable to evaluate client_id parameter")
        return client_id  # type: ignore # value will be returned as a string, or an error will be raised

    def get_client_secret_name(self) -> str:
        return self._client_secret_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_client_secret(self) -> str:
        client_secret = (
            self._client_secret.eval(self.config) if self._client_secret else self._client_secret
        )
        if not client_secret:
            raise ValueError("OAuthAuthenticator was unable to evaluate client_secret parameter")
        return client_secret  # type: ignore # value will be returned as a string, or an error will be raised

    def get_refresh_token_name(self) -> str:
        return self._refresh_token_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_refresh_token(self) -> Optional[str]:
        return None if self._refresh_token is None else str(self._refresh_token.eval(self.config))

    def get_scopes(self) -> List[str]:
        return self.scopes or []

    def get_access_token_name(self) -> str:
        return self.access_token_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_expires_in_name(self) -> str:
        return self.expires_in_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_grant_type_name(self) -> str:
        return self.grant_type_name.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_grant_type(self) -> str:
        return self.grant_type.eval(self.config)  # type: ignore # eval returns a string in this context

    def get_refresh_request_body(self) -> Mapping[str, Any]:
        return self._refresh_request_body.eval(self.config)

    def get_refresh_request_headers(self) -> Mapping[str, Any]:
        return self._refresh_request_headers.eval(self.config)

    def get_token_expiry_date(self) -> AirbyteDateTime:
        if not self._has_access_token_been_initialized():
            return AirbyteDateTime.from_datetime(datetime.min)
        return self._token_expiry_date  # type: ignore # _token_expiry_date is an AirbyteDateTime. It is never None despite what mypy thinks

    def _has_access_token_been_initialized(self) -> bool:
        return self._access_token is not None

    def set_token_expiry_date(self, value: Union[str, int]) -> None:
        self._token_expiry_date = self._parse_token_expiration_date(value)

    def get_assertion_name(self) -> str:
        return self.assertion_name

    def get_assertion(self) -> str:
        if self.profile_assertion is None:
            raise ValueError("profile_assertion is not set")
        return self.profile_assertion.token

    def build_refresh_request_body(self) -> Mapping[str, Any]:
        """
        Returns the request body to set on the refresh request

        Override to define additional parameters
        """
        if self.use_profile_assertion:
            return {
                self.get_grant_type_name(): self.get_grant_type(),
                self.get_assertion_name(): self.get_assertion(),
            }
        return super().build_refresh_request_body()

    @property
    def access_token(self) -> str:
        if self._access_token is None:
            raise ValueError("access_token is not set")
        return self._access_token

    @access_token.setter
    def access_token(self, value: str) -> None:
        self._access_token = value

    @property
    def _message_repository(self) -> MessageRepository:
        """
        Overriding AbstractOauth2Authenticator._message_repository to allow for HTTP request logs
        """
        return self.message_repository


@dataclass
class DeclarativeSingleUseRefreshTokenOauth2Authenticator(
    SingleUseRefreshTokenOauth2Authenticator, DeclarativeAuthenticator
):
    """
    Declarative version of SingleUseRefreshTokenOauth2Authenticator which can be used in declarative connectors.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)



import csv
import gzip
import io
import json
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from io import BufferedIOBase, TextIOWrapper
from typing import Any, Generator, MutableMapping, Optional

import orjson
import requests

from airbyte_cdk.models import FailureType
from airbyte_cdk.sources.declarative.decoders.decoder import Decoder
from airbyte_cdk.utils import AirbyteTracedException

logger = logging.getLogger("airbyte")


@dataclass
class Parser(ABC):
    @abstractmethod
    def parse(
        self,
        data: BufferedIOBase,
    ) -> Generator[MutableMapping[str, Any], None, None]:
        """
        Parse data and yield dictionaries.
        """
        pass


@dataclass
class GzipParser(Parser):
    inner_parser: Parser

    def parse(
        self,
        data: BufferedIOBase,
    ) -> Generator[MutableMapping[str, Any], None, None]:
        """
        Decompress gzipped bytes and pass decompressed data to the inner parser.
        """
        with gzip.GzipFile(fileobj=data, mode="rb") as gzipobj:
            yield from self.inner_parser.parse(gzipobj)


@dataclass
class JsonParser(Parser):
    encoding: str = "utf-8"

    def parse(self, data: BufferedIOBase) -> Generator[MutableMapping[str, Any], None, None]:
        """
        Attempts to deserialize data using orjson library. As an extra layer of safety we fallback on the json library to deserialize the data.
        """
        raw_data = data.read()
        body_json = self._parse_orjson(raw_data) or self._parse_json(raw_data)

        if body_json is None:
            raise AirbyteTracedException(
                message="Response JSON data failed to be parsed. See logs for more information.",
                internal_message=f"Response JSON data failed to be parsed.",
                failure_type=FailureType.system_error,
            )

        if isinstance(body_json, list):
            yield from body_json
        else:
            yield from [body_json]

    def _parse_orjson(self, raw_data: bytes) -> Optional[Any]:
        try:
            return orjson.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.debug(
                f"Failed to parse JSON data using orjson library. Falling back to json library. {exc}"
            )
            return None

    def _parse_json(self, raw_data: bytes) -> Optional[Any]:
        try:
            return json.loads(raw_data.decode(self.encoding))
        except Exception as exc:
            logger.error(f"Failed to parse JSON data using json library. {exc}")
            return None


@dataclass
class JsonLineParser(Parser):
    encoding: Optional[str] = "utf-8"

    def parse(
        self,
        data: BufferedIOBase,
    ) -> Generator[MutableMapping[str, Any], None, None]:
        for line in data:
            try:
                yield json.loads(line.decode(encoding=self.encoding or "utf-8"))
            except json.JSONDecodeError as e:
                logger.warning(f"Cannot decode/parse line {line!r} as JSON, error: {e}")


@dataclass
class CsvParser(Parser):
    # TODO: migrate implementation to re-use file-base classes
    encoding: Optional[str] = "utf-8"
    delimiter: Optional[str] = ","

    def _get_delimiter(self) -> Optional[str]:
        """
        Get delimiter from the configuration. Check for the escape character and decode it.
        """
        if self.delimiter is not None:
            if self.delimiter.startswith("\\"):
                self.delimiter = self.delimiter.encode("utf-8").decode("unicode_escape")

        return self.delimiter

    def parse(
        self,
        data: BufferedIOBase,
    ) -> Generator[MutableMapping[str, Any], None, None]:
        """
        Parse CSV data from decompressed bytes.
        """
        text_data = TextIOWrapper(data, encoding=self.encoding)  # type: ignore
        reader = csv.DictReader(text_data, delimiter=self._get_delimiter() or ",")
        for row in reader:
            yield row


@dataclass
class CompositeRawDecoder(Decoder):
    """
    Decoder strategy to transform a requests.Response into a Generator[MutableMapping[str, Any], None, None]
    passed response.raw to parser(s).
    Note: response.raw is not decoded/decompressed by default.
    parsers should be instantiated recursively.
    Example:
    composite_raw_decoder = CompositeRawDecoder(parser=GzipParser(inner_parser=JsonLineParser(encoding="iso-8859-1")))
    """

    parser: Parser
    stream_response: bool = True

    def is_stream_response(self) -> bool:
        return self.stream_response

    def decode(
        self, response: requests.Response
    ) -> Generator[MutableMapping[str, Any], None, None]:
        if self.is_stream_response():
            yield from self.parser.parse(data=response.raw)  # type: ignore[arg-type]
        else:
            yield from self.parser.parse(data=io.BytesIO(response.content))



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Any, Callable, List, MutableMapping, Optional, Tuple

# FIXME We would eventually like the Concurrent package do be agnostic of the declarative package. However, this is a breaking change and
#  the goal in the short term is only to fix the issue we are seeing for source-declarative-manifest.
from airbyte_cdk.sources.declarative.datetime.datetime_parser import DatetimeParser
from airbyte_cdk.sources.streams.concurrent.cursor import CursorField
from airbyte_cdk.sources.streams.concurrent.state_converters.abstract_stream_state_converter import (
    AbstractStreamStateConverter,
    ConcurrencyCompatibleStateType,
)
from airbyte_cdk.utils.datetime_helpers import AirbyteDateTime, ab_datetime_now, ab_datetime_parse


class DateTimeStreamStateConverter(AbstractStreamStateConverter):
    def _from_state_message(self, value: Any) -> Any:
        return self.parse_timestamp(value)

    def _to_state_message(self, value: Any) -> Any:
        return self.output_format(value)

    @property
    @abstractmethod
    def _zero_value(self) -> Any: ...

    @property
    def zero_value(self) -> datetime:
        return self.parse_timestamp(self._zero_value)

    @classmethod
    def get_end_provider(cls) -> Callable[[], datetime]:
        return ab_datetime_now

    @abstractmethod
    def increment(self, timestamp: datetime) -> datetime: ...

    @abstractmethod
    def parse_timestamp(self, timestamp: Any) -> datetime: ...

    @abstractmethod
    def output_format(self, timestamp: datetime) -> Any: ...

    def parse_value(self, value: Any) -> Any:
        """
        Parse the value of the cursor field into a comparable value.
        """
        return self.parse_timestamp(value)

    def _compare_intervals(self, end_time: Any, start_time: Any) -> bool:
        return bool(self.increment(end_time) >= start_time)

    def convert_from_sequential_state(
        self,
        cursor_field: CursorField,
        stream_state: MutableMapping[str, Any],
        start: Optional[datetime],
    ) -> Tuple[datetime, MutableMapping[str, Any]]:
        """
        Convert the state message to the format required by the ConcurrentCursor.

        e.g.
        {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "metadata": { … },
            "slices": [
                {"start": "2021-01-18T21:18:20.000+00:00", "end": "2021-01-18T21:18:20.000+00:00"},
            ]
        }
        """
        sync_start = self._get_sync_start(cursor_field, stream_state, start)
        if self.is_state_message_compatible(stream_state):
            return sync_start, stream_state

        # Create a slice to represent the records synced during prior syncs.
        # The start and end are the same to avoid confusion as to whether the records for this slice
        # were actually synced
        slices = [
            {
                self.START_KEY: start if start is not None else sync_start,
                self.END_KEY: sync_start,
                self.MOST_RECENT_RECORD_KEY: sync_start,
            }
        ]

        return sync_start, {
            "state_type": ConcurrencyCompatibleStateType.date_range.value,
            "slices": slices,
            "legacy": stream_state,
        }

    def _get_sync_start(
        self,
        cursor_field: CursorField,
        stream_state: MutableMapping[str, Any],
        start: Optional[datetime],
    ) -> datetime:
        sync_start = start if start is not None else self.zero_value
        prev_sync_low_water_mark = (
            self.parse_timestamp(stream_state[cursor_field.cursor_field_key])
            if cursor_field.cursor_field_key in stream_state
            else None
        )
        if prev_sync_low_water_mark and prev_sync_low_water_mark >= sync_start:
            return prev_sync_low_water_mark
        else:
            return sync_start


class EpochValueConcurrentStreamStateConverter(DateTimeStreamStateConverter):
    """
    e.g.
    { "created": 1617030403 }
    =>
    {
        "state_type": "date-range",
        "metadata": { … },
        "slices": [
            {starts: 0, end: 1617030403, finished_processing: true}
        ]
    }
    """

    _zero_value = 0

    def increment(self, timestamp: datetime) -> datetime:
        return timestamp + timedelta(seconds=1)

    def output_format(self, timestamp: datetime) -> int:
        return int(timestamp.timestamp())

    def parse_timestamp(self, timestamp: int) -> datetime:
        dt_object = AirbyteDateTime.fromtimestamp(timestamp, timezone.utc)
        if not isinstance(dt_object, AirbyteDateTime):
            raise ValueError(
                f"AirbyteDateTime object was expected but got {type(dt_object)} from AirbyteDateTime.fromtimestamp({timestamp})"
            )
        return dt_object


class IsoMillisConcurrentStreamStateConverter(DateTimeStreamStateConverter):
    """
    e.g.
    { "created": "2021-01-18T21:18:20.000Z" }
    =>
    {
        "state_type": "date-range",
        "metadata": { … },
        "slices": [
            {starts: "2020-01-18T21:18:20.000Z", end: "2021-01-18T21:18:20.000Z", finished_processing: true}
        ]
    }
    """

    _zero_value = "0001-01-01T00:00:00.000Z"

    def __init__(
        self, is_sequential_state: bool = True, cursor_granularity: Optional[timedelta] = None
    ):
        super().__init__(is_sequential_state=is_sequential_state)
        self._cursor_granularity = cursor_granularity or timedelta(milliseconds=1)

    def increment(self, timestamp: datetime) -> datetime:
        return timestamp + self._cursor_granularity

    def output_format(self, timestamp: datetime) -> str:
        """Format datetime with milliseconds always included.

        Args:
            timestamp: The datetime to format.

        Returns:
            str: ISO8601/RFC3339 formatted string with milliseconds.
        """
        dt = AirbyteDateTime.from_datetime(timestamp)
        # Always include milliseconds, even if zero
        millis = dt.microsecond // 1000 if dt.microsecond else 0
        return f"{dt.year:04d}-{dt.month:02d}-{dt.day:02d}T{dt.hour:02d}:{dt.minute:02d}:{dt.second:02d}.{millis:03d}Z"

    def parse_timestamp(self, timestamp: str) -> datetime:
        dt_object = ab_datetime_parse(timestamp)
        if not isinstance(dt_object, AirbyteDateTime):
            raise ValueError(
                f"AirbyteDateTime object was expected but got {type(dt_object)} from parse({timestamp})"
            )
        return dt_object


class CustomFormatConcurrentStreamStateConverter(IsoMillisConcurrentStreamStateConverter):
    """
    Datetime State converter that emits state according to the supplied datetime format. The converter supports reading
    incoming state in any valid datetime format using AirbyteDateTime parsing utilities.
    """

    def __init__(
        self,
        datetime_format: str,
        input_datetime_formats: Optional[List[str]] = None,
        is_sequential_state: bool = True,
        cursor_granularity: Optional[timedelta] = None,
    ):
        super().__init__(
            is_sequential_state=is_sequential_state, cursor_granularity=cursor_granularity
        )
        self._datetime_format = datetime_format
        self._input_datetime_formats = input_datetime_formats if input_datetime_formats else []
        self._input_datetime_formats += [self._datetime_format]
        self._parser = DatetimeParser()

    def output_format(self, timestamp: datetime) -> str:
        return self._parser.format(timestamp, self._datetime_format)

    def parse_timestamp(self, timestamp: str) -> datetime:
        for datetime_format in self._input_datetime_formats:
            try:
                return self._parser.parse(timestamp, datetime_format)
            except ValueError:
                pass
        raise ValueError(f"No format in {self._input_datetime_formats} matching {timestamp}")



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Optional

from airbyte_cdk.sources.types import Config, StreamSlice, StreamState


@dataclass
class RecordTransformation:
    """
    Implementations of this class define transformations that can be applied to records of a stream.
    """

    @abstractmethod
    def transform(
        self,
        record: Dict[str, Any],
        config: Optional[Config] = None,
        stream_state: Optional[StreamState] = None,
        stream_slice: Optional[StreamSlice] = None,
    ) -> None:
        """
        Transform a record by adding, deleting, or mutating fields directly from the record reference passed in argument.

        :param record: The input record to be transformed
        :param config: The user-provided configuration as specified by the source's spec
        :param stream_state: The stream state
        :param stream_slice: The stream slice
        :return: The transformed record
        """

    def __eq__(self, other: object) -> bool:
        return other.__dict__ == self.__dict__
